/*
 * Copyright Bigeon Emmanuel (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * provide a generic framework for console applications.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
/** gclc:fr.bigeon.gclc.ConsoleApplication.java
 * Created on: Sep 6, 2014 */
package fr.bigeon.gclc;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.command.ExitCommand;
import fr.bigeon.gclc.command.HelpExecutor;
import fr.bigeon.gclc.command.ICommand;
import fr.bigeon.gclc.command.ICommandProvider;
import fr.bigeon.gclc.command.SubedCommand;
import fr.bigeon.gclc.command.UnrecognizedCommand;
import fr.bigeon.gclc.exception.CommandParsingException;
import fr.bigeon.gclc.exception.CommandRunException;
import fr.bigeon.gclc.exception.InvalidCommandName;
import fr.bigeon.gclc.i18n.Messages;
import fr.bigeon.gclc.manager.ConsoleManager;
import fr.bigeon.gclc.manager.SystemConsoleManager;

/** <p>
 * A {@link ConsoleApplication} is an application that require the user to input
 * commands.
 * <p>
 * A typical use case is the following:
 *
 * <pre>
 * {@link ConsoleApplication} app = new {@link ConsoleApplication#ConsoleApplication(String, String, String) ConsoleApplication("exit", "welcome", "see you latter")};
 * app.{@link ConsoleApplication#add(ICommand) add}("my_command", new {@link ICommand MyCommand()});
 * app.{@link ConsoleApplication#start start}();
 * </pre>
 * <p>
 * That will launch in the console application that will display "welcome",
 * return a line and prompt the user for the command until he inputs the
 * <code>exit</code> command. Then it will print "see you latter" and exit the
 * start method.
 *
 * @author Emmanuel BIGEON */
public class ConsoleApplication implements ICommandProvider {

    /** The class logger */
    private static final Logger LOGGER = Logger
            .getLogger(ConsoleApplication.class.getName());
    /** The welcome message */
    private final String header;
    /** The good bye message */
    private final String footer;
    /** The console manager */
    protected final ConsoleManager manager;
    /** The container of commands */
    private final SubedCommand root;
    /** The state of this application */
    private boolean running;
    /** The listeners */
    private final List<CommandRequestListener> listeners = new ArrayList<>();

    /** @param manager the manager
     * @param welcome the welcoming message
     * @param goodbye the goodbye message */
    public ConsoleApplication(ConsoleManager manager, String welcome,
            String goodbye) {
        this.header = welcome;
        this.footer = goodbye;
        this.manager = manager;
        root = new SubedCommand(new String(), new UnrecognizedCommand(manager));
    }

    /** @param manager the manager
     * @param exit the keyword for the exit command of this application
     * @param welcome the header message to display on launch of this
     *            application
     * @param goodbye the message to display on exit
     * @throws InvalidCommandName if the exit command name is invalid
     * @deprecated since 1.2.0, use the {@link #add(ICommand)} method to add the
     *             exit command */
    @Deprecated
    public ConsoleApplication(ConsoleManager manager, String exit,
            String welcome, String goodbye) throws InvalidCommandName {
        this(manager, welcome, goodbye);
        root.add(new ExitCommand(exit, this));
    }

    /** @param exit the keyword for the exit command of this application
     * @param welcome the header message to display on launch of this
     *            application
     * @param goodbye the message to display on exit
     * @throws InvalidCommandName if the exit command name is invalid
     * @deprecated since 1.2.0, use the {@link #add(ICommand)} method to add the
     *             command and create this console using the
     *             {@link #ConsoleApplication(ConsoleManager, String, String)}
     *             method with a {@link SystemConsoleManager} as argument */
    @Deprecated
    public ConsoleApplication(String exit, String welcome,
            String goodbye) throws InvalidCommandName {
        this(new SystemConsoleManager(), welcome, goodbye);
        root.add(new ExitCommand(exit, this));
    }

    @Override
    public final boolean add(ICommand cmd) throws InvalidCommandName {
        return root.add(cmd);
    }

    /** Adds help command on the given key
     *
     * @param cmd the handle for help
     * @return if the help command was added
     * @throws InvalidCommandName if the help command was not valid
     * @deprecated since 1.2.0 use the {@link #add(ICommand)} with a
     *             {@link HelpExecutor} instead */
    @Deprecated
    public final boolean addHelpCommand(String cmd) throws InvalidCommandName {
        return root.add(new HelpExecutor(cmd, manager, root));
    }

    /** Add a command request listener.
     * <p>
     * A listener can listen several times to the same application.
     * 
     * @param listener the listener to add. */
    public final void addListener(CommandRequestListener listener) {
        listeners.add(listener);
    }

    /* (non-Javadoc)
     * @see
     * fr.bigeon.gclc.command.ICommandProvider#executeSub(java.lang.String,
     * java.lang.String[]) */
    @Override
    public final void executeSub(String command,
                                 String... args) throws CommandRunException {
        root.executeSub(command, args);
    }

    /** Exit this running application before next command prompt */
    public final void exit() {
        LOGGER.fine("Request exiting application..."); //$NON-NLS-1$
        running = false;
    }

    /* (non-Javadoc)
     * @see fr.bigeon.gclc.command.ICommandProvider#get(java.lang.String) */
    @Override
    public final ICommand get(String command) {
        return root.get(command);
    }

    /** @return the manager */
    public final ConsoleManager getManager() {
        return manager;
    }

    /** @param cmd the command to interpret
     * @throws IOException if the manager was closed */
    public final void interpretCommand(String cmd) throws IOException {
        List<String> args;
        try {
            args = GCLCConstants.splitCommand(cmd);
        } catch (CommandParsingException e1) {
            manager.println("Command line cannot be parsed"); //$NON-NLS-1$
            LOGGER.log(Level.INFO, "Invalid user command " + cmd, e1); //$NON-NLS-1$
            return;
        }
        if (!args.isEmpty()) {
            try {
                executeSub(args.get(0), Arrays.copyOfRange(
                        args.toArray(new String[0]), 1, args.size()));
            } catch (final CommandRunException e) {
                LOGGER.log(Level.WARNING, "Command failed: " + cmd, e); //$NON-NLS-1$
                manager.println(Messages
                        .getString("ConsoleApplication.cmd.failed", cmd)); //$NON-NLS-1$
                manager.println(e.getLocalizedMessage());
            }
        }
    }

    /** @param listener the listener to remove (once) */
    public final void removeListener(CommandRequestListener listener) {
        for (int i = 0; i < listeners.size(); i++) {
            if (listeners.get(i) == listener) {
                listeners.remove(i);
                return;
            }
        }
    }

    /** Launches the prompting application */
    public final void start() {
        try {
            running = true;
            if (header != null) {
                manager.println(header);
            }
            do {
                final String cmd = manager.prompt();
                if (cmd == null || cmd.isEmpty()) {
                    continue;
                }
                for (final CommandRequestListener listener : listeners) {
                    listener.commandRequest(cmd);
                }
                interpretCommand(cmd);
            } while (running);
            if (footer != null) {
                manager.println(footer);
            }
            LOGGER.fine("Exiting application."); //$NON-NLS-1$
        } catch (IOException e) {
            // The manager was closed
            running = false;
            LOGGER.log(Level.WARNING,
                    "The console manager was closed. Closing the application as no one can reach it.", //$NON-NLS-1$
                    e);
        }
    }

    /** @return if the application is running */
    public boolean isRunning() {
        return running;
    }

    /** @return the root */
    public SubedCommand getRoot() {
        return root;
    }
}
